/*****
*
* ATBAT.PRG
*
* 16 June 91
* David Morgan
*
* Copyright (c) 1991 Nantucket Corporation.
*                    All Rights Reserved.
*
* Clipper Atbat /N/W/A
* RTLINK FILE Atbat PLL base50
*
* Requires Clipper v5.01
*
*   Illustrates retry dialog
*

Illustrates interplay between an application subsystem and an error
handler during a retry dialog.  The error object is the protocol
carrier by which the subsystem and handler communicate with each
other.

Models a hypothetical baseball at-bat.  The at-bat itself has a set
of outcomes.  Success of an at-bat requires one of its conclusive
outcomes, such as a hit or putout.  The atbat begins with a pitch,
which has 1 of 3 outcomes:  strike, ball, or contact.  Contact can be
further processed to yield a hit or putout.  But strikes and balls
cannot, so cannot lead to at-bat success.  They are therefore
classified as erroneous and treated by the error system.

The subsystem submits the strike/ball to the error handler if not yet
the 3rd strike or 4th ball, with invitation to retry.  A 3rd strike
or 4th ball is distinguished as such and submitted, without retry
offer.

The error handler is written to take up offers of retry.  When it
does, the subsystem as (deliberately) written loops back for another
pitch.  When not offered retry, the handler calls its predecessor
error block in any chain.  It adopts the expectation that the
earliest error block will follow the convention to BREAK.  Calling up
the chain is therefore sufficient to assure eventual execution of any
application-specific RECOVER code that the subsystem might contain.
In this example that can be achieved by modifying standare 5.01
ERRORSYS.PRG and linking in the compiled result.  Un-comment line 73,
and comment line 74, to add BREAK to the possible outcomes of
DefError().

This illustration uses color conventions for messages, according to
source.  All messages from the subsystem are light blue on blue; from
the handler, yellow on red.  From the standard DefError() they are
white on red.

To view messages from the handler SET ATBAT=SHOW.  To force all
pitches to strikes or balls, incorporate keywords ALLSTRIKES or
ALLBALLS in the ATBAT environment variable.

See related articles in Nantucket News "Errors of Your Own Making:
Error Processing, Theory and Practice," Part I, vol 5 no 2
September/October 1990, Part II, vol 5 no 4 January/February 1991.
*/

#include "Box.ch"
#include "Setcurs.ch"

#define BALLS          1
#define STRIKES        2

FUNCTION Main
	LOCAL cOldColor := SETCOLOR("BG+/B")
   LOCAL nOldCurs  := SETCURSOR(SC_NONE)
   LOCAL cResult

   CLEAR SCREEN

   cResult := AtBat()
   @ 5,5,11,25 BOX B_SINGLE
   SETCOLOR("BG+/B*")
   @ 8,12 SAY cResult
   INKEY(0)

   SETCURSOR(nOldCurs)
   SETCOLOR(cOldColor)
   CLEAR SCREEN

   RETURN (NIL)


FUNCTION AtBat
   STATIC oOldBlock
   LOCAL aCount := {0,0}
   LOCAL oErr
   LOCAL cPitch
   LOCAL uResult
   LOCAL lRetry := .T.
   LOCAL cRetVal
   LOCAL oWhatHappened

   oOldBlock := ERRORBLOCK({|e| MyHandler(e, oOldBlock)})

   BEGIN SEQUENCE

      oErr := GenErr()  // build foundation-level err obj

      DO WHILE lRetry

         cPitch := Pitch()

         lRetry := .T.
         IF .NOT. (cPitch == "Contact")
            aCount[ASCAN({"Ball", "Strike"}, cPitch)]++
            oErr:tries++
            IF (aCount[STRIKES] < 3) .AND. (aCount[BALLS] < 4)
               oErr:description := cPitch
               oErr:canRetry := .T.
               ShowCount(aCount)
            ELSEIF aCount[STRIKES] == 3
               oErr:description := "Third Strike"
               oErr:canRetry := .F.
            ELSEIF aCount[BALLS] == 4
               oErr:description := "Fourth Ball"
               oErr:canRetry := .F.
            ENDIF

            // Raise error, capture possible feedback
            lRetry := EVAL(ERRORBLOCK(), oErr) 

            IF lRetry
               LOOP                 // causes retry; handler asked you to
            ENDIF

         ELSE
            lRetry := .F.           // no violation, no retry
         ENDIF

         //   <continuation of operation>
         cRetVal := FlyGround()

      ENDDO

   RECOVER USING oWhatHappened

      IF oWhatHappened:description == "Third Strike"
         cRetVal := "Strikeout"
      ELSEIF oWhatHappened:description == "Fourth Ball"
         cRetVal := "Walk"
      ELSE
         BREAK(oWhatHappened)    // no situation known to me, particularly...
      ENDIF

   END // SEQUENCE

   ERRORBLOCK(oOldBlock)
   RETURN (cRetVal)


STATIC FUNCTION MyHandler( oObj, bPredecessor )

   LOCAL cOldColor := SETCOLOR("GR+/R+")
   LOCAL cOldScreen := SAVESCREEN(0, 0, 24, 79)

   IF "SHOW" $ UPPER(GETENV("ATBAT"))
      ShowObject(oObj)
   ENDIF

   IF oObj:subSystem == "AtBat" .AND. ;
         (oObj:description=="Strike" .OR. oObj:description=="Ball") .AND. ;
         oObj:canRetry

      IF "SHOW" $ UPPER(GETENV("ATBAT"))
         DeclareIntent("Will now take up the offered option to request a retry.")
      ENDIF
      SETCOLOR(cOldColor); RESTSCREEN(0, 0, 24, 79, cOldScreen)
      RETURN (.T.)

   ELSEIF oObj:subSystem == "AtBat" .AND.;
      (oObj:description=="Third Strike" .OR. oObj:description=="Fourth Ball") ;
       .AND. .NOT. oObj:canRetry

      IF "SHOW" $ UPPER(GETENV("ATBAT"))
         DeclareIntent("No option to retry.  Don't wish to QUIT.  "+ ;
            "Two choices left:  BREAK or EVAL() predecessor.  "+ ;
            "Will defer to predecessor, relying by convention on "+ ;
            "an eventual BREAK.  Gives both predecessor(s) and "+ ;
            "local routine's RECOVERy code (if any) a crack.")
      ENDIF
      SETCOLOR(cOldColor); RESTSCREEN(0, 0, 24, 79, cOldScreen)
   ENDIF

   RETURN (EVAL(bPredecessor, oObj))


STATIC FUNCTION ShowCount( aCount )
   LOCAL cOldScreen := SAVESCREEN(5, 5, 11, 25)

   @ 5,5 CLEAR TO 11,25
   @ 5,5,11,25 BOX B_SINGLE
   @ 7,11 SAY "Ball "+ LTRIM(STR(aCount[BALLS]))
   @ 8,11 SAY "Strike "+ LTRIM(STR(aCount[STRIKES]))
   @ 10,7 SAY "Any Key Continues"
   INKEY(0)
   RESTSCREEN(5, 5, 11, 25, cOldScreen)

   RETURN (NIL)


STATIC FUNCTION Pitch
   STATIC aOutcomes := {"Ball", "Ball", "Strike", "Contact"}
   LOCAL cPitch, i, j

   IF "ALLBALLS" $ UPPER(GETENV("ATBAT"))
      cPitch := "Ball"
   ELSEIF "ALLSTRIKES" $ UPPER(GETENV("ATBAT"))
      cPitch := "Strike"
   ELSE
      Randomize()
      cPitch := aOutcomes[RndInt(4)]
   ENDIF

   // Screen evidence of outcome
   j := 10
   FOR i := 78 TO 1 STEP -1
      @ j,i+1 SAY ' '
      @ j += IIF(cPitch=="Ball".AND.i%10==0,1,0),i SAY CHR(7)
      IIF(i % 10 == 0, INKEY(.1), NIL)
   NEXT
   @ j,i+1 SAY ' '
   IIF(cPitch == "Contact", StarBurst(), NIL)

   RETURN (cPitch)


STATIC FUNCTION StarBurst
   LOCAL c := COL()
   LOCAL cDisp
   LOCAL k := 1
   LOCAL n
   LOCAL p
   LOCAL r := ROW()
   LOCAL s
   LOCAL t
   LOCAL aShift := { - 1, 0, -2, 0, -3, 0, -1, 1, -2, 2, -3, 3, 0, 0, 0, 2, 0,;
      4, 1, 1, 2, 2, 3, 3, 1, 0, 2, 0, 3, 0}

   FOR n := 1 TO 50
      FOR t := 0 TO 4
         FOR p := 1 TO 2
            cDisp := IIF(p == 1, CHR(7), " ")
            FOR s := 1 TO 3
               @ r+aShift[k++],c+aShift[k++] SAY cDisp
            NEXT
            k -= IIF(p == 1, 6, 0)
         NEXT
      NEXT
      k := 1
   NEXT

   RETURN (NIL)


STATIC FUNCTION FlyGround
   STATIC aOutComes := {"Out", "Out", "Out", "Single", "Single", "Double", "Triple", "Homerun"}
   Randomize()
   RETURN (aOutcomes[RndInt(8)])


FUNCTION Randomize
   // Randomly alter value of RndInt()'s seed prior to using RndInt()
   LOCAL i
   LOCAL Max := 100 * (SECONDS() - INT(SECONDS()))

   FOR i := 1 TO Max
      RndInt(1)
   NEXT
   RETURN (NIL)


FUNCTION RndInt( nCeiling )
   // Generates random integer where  1 < value <= nCeiling
   STATIC nSeed := 0.123456789
   nSeed *= 997
   nSeed -= INT(nSeed)

   RETURN (INT(nCeiling * nSeed) + 1)


STATIC FUNCTION GenErr
   LOCAL oNewError
   oNewError := ErrorNew()
   oNewError:canDefault := .F.
   oNewError:canSubstitute := .F.
   oNewError:subSystem := "AtBat"
   oNewError:tries := 0

   RETURN (oNewError)


/***
*   This static function displays the variables associated with the
*   current Error Object.
*
* Originated by Philip Schwartz
*/
STATIC FUNCTION ShowObject( e )
   LOCAL i

   CLS
   ? "Current Error Object Variables:"
   ? REPLICATE( "-", 80 )
   ? "e:canDefault is:           " + IF( e:canDefault, "TRUE", "FALSE" )
   ? "e:canRetry is:             " + IF( e:canRetry, "TRUE", "FALSE" )
   ? "e:canSubstitute is:        " + IF( e:cansubstitute,"TRUE", "FALSE" )
   ? "e:description() is:        " + e:description()
   ? "e:filename() is:           " + e:filename()
   ? "e:genCode() is:            " + LTRIM( STR( e:genCode() ) )
   ? "e:operation() is:          " + e:operation()
   ? "e:osCode() is:             " + LTRIM( STR( e:osCode() ) )
   ? "e:subCode() is:            " + LTRIM( STR( e:subCode() ) )
   ? "e:subsystem() is:          " + e:subsystem()
   ? "e:tries() is:              " + LTRIM( STR( e:tries() ) )
   ? "e:severity() is:           " + LTRIM( STR( e:severity() ) )
   ? "e:args() is:               "

   IF VALTYPE(e:args()) == "A"
      ?? "An ARRAY: " + LTRIM( STR( LEN( e:args() ) ) ) + " Element(s)"

      FOR i := 1 TO LEN(e:args())
         ? "e:args[" + STR(i, 2) + "] is:          ", e:args[i]
      NEXT

   ELSE
      ?? "Not available"
   ENDIF

   INKEY(0)
   RETURN (NIL)


STATIC PROCEDURE DeclareIntent( cMessage )
   LOCAL lOldScoreBoard := SET(_SET_SCOREBOARD, .F.)
   @ 18,0,24,79 BOX B_SINGLE
   MEMOEDIT(cMessage, 19, 1, 23, 78, .F.)
   SET(_SET_SCOREBOARD, lOldScoreBoard)
   RETURN

// EOF: AtBat.prg
