/*****
*
* TRANSACT.PRG
*
* 29 April 91
* David Morgan
*
* Copyright (c) 1991 Nantucket Corporation.
*                    All Rights Reserved.
*
* Clipper Transact /N/W/A
* RTLINK FILE Transact,Locks PLL base50
*
* Requires Clipper v5.01
*
*/
#include "error.ch"

PROCEDURE Main
/*
*	Illustrates use of Clipper 5.0 error features
*
*		1) to implement the transaction/rollback concept
*		2) to restart/retry a block of statements
*
	The illustrated transaction supposes a One-to-Many relationship
	between two data files.  The transaction is defined to consist of
	applying DELETEs to all of the Many (children), as a group, that
	correspond to a given One (parent).  In this case, all the invoices
	associated with customer Jones.  Rollback is defined to be the
	application of RECALLs, in case any target record was not DELETEd, to
	those target records that were.  (The concurrency implications of
	these definitions may or may not be adequate for particular
	purposes.)  The anticipated obstacle, for which rollback is provided,
	is unRLOCKability of one of the target records on a network.

	The installed error block, though minimalistic, embodies the
	recommendation that error block code eventually lead to a BREAK so
	that local code can be accompanied by local recovery specific to
	local activity, and all errors will reliably transfer control to that
	recovery.  (The error block shown here accomplishes this
	artificially.  More generally, it is the highest-level block in a
	possible chain that would perform the BREAK, upon which this code
	would rely.  See Nantucket News, Jan/Feb 1991, vol 5  no 4.)

	Note that there is RECOVERy with the start-up code (GetReady) for
	the start-up code, and separately, RECOVERy with the multiple-delete
	code for the multiple-delete code.  If an error arises in either
	context, execution of the localized RECOVERy code is thanks in both
	cases to one and the same error block.  (Once posted, the error block
	does not "belong" to the routine that posted it, but to any and all
	routines that execute while it remains installed.)
*/

	LOCAL aDeleteds, cName, aNotRecalled, oWhatHappened
	FIELD LastName
	ERRORBLOCK( {|e| BREAK(e)} )  // Artificial but prototypical
	IF .NOT.GetReady()  // Open files, Set relation
		? "No I don't feel like it tonight"; QUIT
	ENDIF

  IF .NOT.Invoice->(DBSEEK (cName := "Jones"))
		? 'Oh come on.  This example is hard-wired for the "Joneses."'; QUIT
	ENDIF

	IF Customer->(RecLock(5))  // parent record lock, gateway to its children

		DO WHILE .T.                      // allow for possible retry

			BEGIN SEQUENCE

				//  Transaction Attempt (multiple DELETEs)
				aDeleteds := {}
				DO WHILE TRIM(Invoice->(LastName))=="Jones"
					Invoice->(RLOCK())         // may not succeed
					Invoice->(DBDELETE())      // in that case, this calls error block
					AADD(aDeleteds, Invoice->(RECNO()))  // records "damage" already done
					Invoice->(DBSKIP())
				ENDDO

			RECOVER USING oWhatHappened

				IF (oWhatHappened:genCode() == EG_UNLOCKED)

					//  Rollback (UnTransact) attempt
					DO WHILE .NOT.EMPTY(aDeleteds)        // stay till they're squashed
						aNotRecalled := {}
						AEVAL(aDeleteds, { |RecNum| UnDeleteIt(RecNum, aNotRecalled) } )
						aDeleteds := ACLONE(aNotRecalled)
					ENDDO

					IF (ErrBox("Rollback successful, Transaction NOT Performed; " + "Restart transaction again?",{"Yes","No"}) == 1)
						Invoice->(DBSEEK(cName := "Jones"))     // reposition pointer
						LOOP                                    // ...and start over
					ENDIF

				ENDIF

			END SEQUENCE
			EXIT

		ENDDO

		Customer->(DBUNLOCK())
	ENDIF

	RETURN


STATIC PROCEDURE UnDeleteIt(RecNum, aNotRecalled)
	Invoice->(DBGOTO(RecNum))
	IF Invoice->(RLOCK())                      // either...
		Invoice->(DBRECALL())                    //  uneventful RECALL
		Invoice->(DBUNLOCK())
	ELSE
		AADD(aNotRecalled, RecNum)               //  ... or mark for revisit
	ENDIF
	RETURN


STATIC FUNCTION GetReady
/*
	Illustrates use of error system to treat network file "acquisition"
	as a transaction.  Dispenses with Summer '87 NETERR() check after USE
	attempt as a result.  Streamlines code and makes it more readable.

	Need to accomplish 3-steps-in-1, or nothing:  open 2 files and
	indexes, and set a relation.  That defines the transaction.  Relies
	on knowledge that error block will predictably BREAK.  Therefore if
	any part of the main sequecnce of statements fails the RECOVER
	statement block that gets control can undo the parts that did not.
	RETURN value reflects overall result.
*/

	LOCAL lReturn := .T.
	CLEAR SCREEN
	BEGIN SEQUENCE
		USE Invoice  INDEX Invoice  SHARED NEW
		USE Customer INDEX Customer SHARED NEW
		SET RELATION TO Customer->LastName INTO Invoice
	RECOVER		// for ANY error; don't grab object, don't check why
 		CLOSE SELECT("Invoice")
		CLOSE SELECT("Customer")
		lReturn := .F.
	END SEQUENCE
	RETURN lReturn



FUNCTION ErrBox( cErrMsg, aOptions )
*	originated by Philip H. Schwartz
LOCAL nChoice, i, nStrgLen, cOldColor, cOldScreen

cOldColor := SETCOLOR( "W/R+, W+/B" )
cOldScreen := SAVESCREEN(10,0,14,79)

@ 10, 0 TO 14, 79 DOUBLE
@ 11, 1, 13, 78 BOX SPACE( 9 )

@ 11, 2 SAY cErrMsg

nStrgLen := LEN( aOptions ) * 3

AEVAL( aOptions, { |arr| nStrgLen += LEN( arr ) } )

nStrgLen := (78 - nStrgLen ) / 2

FOR i := 1 TO LEN( aOptions )
  @ 13, nStrgLen PROMPT aOptions[i]
  nStrgLen += LEN( aOptions[i] ) + 3
NEXT

nChoice := 1
MENU TO nChoice

SETCOLOR( cOldColor )
RESTSCREEN(10,0,14,79,cOldScreen)

RETURN( nChoice )


// EOF - TRANSACT.PRG //
